/*	$LAAS$ */
/*
 * Copyright (c) 2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#ifndef JLOCO_PRIVATE_H
#define JLOCO_PRIVATE_H

#include "jlocoStruct.h"
#include "odoProba.h"
#include <posterLib.h>
#include <genPosStruct.h>


#define PDEBUG_ERR(fmt, args...) fprintf(stderr, "%s(): "fmt, __func__, ## args)
#define PDEBUG PDEBUG_ERR

#ifdef DEBUG
# define PVDEBUG PDEBUG_ERR
#else
# define PVDEBUG(fmt, args...) /* nothing */
#endif


#ifdef PLOTDEBUG
# define PPLOT(fmt, args...) fprintf(stderr, "PLOT: "fmt, ## args)
#else
# define PPLOT(fmt, args...) /* nothing */
#endif


#ifndef max
# define max(x,y) ((x)>(y)?(x):(y))
#endif



typedef struct JLOCO_PRIVATE {
  JLOCO_MODE        trackMode;
  double            cmdTaskPeriod;
  GENPOS_CART_REF   posRef;           // trajectory tracking reference
  GENPOS_CART_SPEED speedRef;         // speed tracking reference
  POSTER_ID         trackedPoster;
  
  double motRr_nom;		// nominal radius of right wheel
  double motRl_nom;		// nominal radius of left wheel
  double motW_nom;		// nominal distance between wheels
  double motRr_cor;		// corrected radius of right wheel
  double motRl_cor;		// corrected radius of left wheel
  double motW_cor;		// corrected distance between wheels
  
  double sdMax;			// max translational speed
  double thetadMax;		// max rotational speed
  double sddMax;		// max translational acceleration
  double thetaddMax;		// max rotational acceleration
  
  int waitRefCounter;           // for external watchdog
  int cmdMismatchThreshold;     // too many speed commands ignored?
  int pumpMismatchThreshold;    // pump->cmd watchdog failed?
  int importantDrift;           // detect position tracking problems
  
  JLOCO_ODO odo;	        // odometry parameters, cache, and values
} JLOCO_PRIVATE;


extern JLOCO_PRIVATE *jlocoPrivate;


/** Fill jp->odo and jp->gyroOdo according to the provided
    parameters. This does NOT touch the following fields:
    robotMisc.{s,cR,dTheta,dS} robot.{v,w} {right,left}Wheel. The
    positional uncertainty is initialized to a diagonal covariance
    matrix with the given coefficients vxx, vyy, vtt.

    \note This function assumes the robot is moving forwards for
    computing the "control point".
*/
void initOdometry(/** structure to be filled, typically jlocoPrivate */
		  JLOCO_PRIVATE * jp,
		  /** global x-position */
		  double x,
		  /** global y-position */
		  double y,
		  /** global heading (theta-position) */
		  double theta,
		  /** uncertainty in x */
		  double vxx,
		  /** uncertainty in y */
		  double vyy,
		  /** uncertainty in theta */
		  double vtt,
		  /** distance of "controlled point" from robot center */
		  double cntrl_dist);


//////////////////////////////////////////////////
// FAKE_NEOBOTIX declarations for debugging and offline development

#ifdef FAKE_NEOBOTIX

#define CAN_DEVICE_PATH "/dev/null"
#define CAN_IO_EMERGENCY_STOP 0x1000

typedef struct {
  double lvel;			// left wheel speed [rad/s]
  double rvel;			// right wheel speed [rad/s]
  double gphi;			// gyro angle [rad]
  double dt;			// timestep [s]
} NEO_STR;

typedef enum CAN_MOTOR_STATE {
  ST_PRE_INITIALIZED,
  ST_OPERATION_ENABLED,
  ST_OPERATION_DISABLED,
  ST_MOTOR_FAILURE
} CAN_MOTOR_STATE;

typedef enum CAN_MOTOR_ID {
  CAN_MOTOR_LEFT,
  CAN_MOTOR_RIGHT
} CAN_MOTOR_ID;

static NEO_STR *neoInit(char *dev);
static void neoEnd(NEO_STR *neo);
static STATUS canMotorsStart(NEO_STR *neo);
static STATUS canMotorsReset(NEO_STR *neo);
static CAN_MOTOR_STATE canMotorsState(NEO_STR *neo, CAN_MOTOR_ID id);
static void canMotorsGetGearDeltaPosVelRadS(NEO_STR *neo,
					    double *lDeltaPos, double *lvel,
					    double *rDeltaPos, double *rvel);
static void canMotorsSetVelGearRadS(NEO_STR *neo, double lvel, double rvel);
static void canGyroRequestConnect(NEO_STR *neo);
static void canGyroZero(NEO_STR *neo);
static void canGyroRequestAngle(NEO_STR *neo);
static double canGyroGetAngle(NEO_STR *neo);
static int canIOGetVBatt(NEO_STR *neo);
static unsigned short canIOGetDigiIn(NEO_STR *neo);
static void canIORequestVBatt(NEO_STR *neo);
static void canIORequestDigiIn(NEO_STR *neo);

#endif // FAKE_NEOBOTIX

#endif // JLOCO_PRIVATE_H
