/*
 * Copyright (c) 2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/*--------------------------------------------------------------------------
 *    
 *                         -- C N R S -- 
 *         Laboratoire d'Automatique et d'Analyse des Systemes 
 *                 7 Avenue du colonel Roche   
 *                     31 077 Toulouse Cedex
 *
 *  Auteur           : Sara Fleury
 *  Groupe           : Robotique et Intelligence Artificielle
 *  Nom du fichier   : piloCmdTaskFunc.c
 *  Date de creation : Juin 1992
 * 
 *  Description      : Calcul de l'odometrie "probabiliste"
 *                     Deux techniques sont proposees:
 *                     odoProba0: Philippe Moutarlier
 *                     odoProba:  Christophe Dousson
 *
 *  Copyright (C) 1992 LAAS/CNRS.
 *----------------------------------------------------------------------*/
/* $LAAS$ */
/*************************************************************************
  Entree: . Deplacements elementaires des roues dR, dL fournis par les codeurs.
          . Position precedente globale (x,y,t) avec la matrice de covariances.
 
  Sortie: . Position dans le repere absolu.
          . Covariance sur cette position dans ce meme repere.
*/

#include <stdio.h>
#include <stdlib.h>
#include <portLib.h>

#include <math.h>
#include "h2mathLib.h"
#include "genPosStruct.h"
#include "jlocoStruct.h"
#include "jlocoConst.h"

#include "odoProba.h"

/* jacobiennes (J_h_Xr,J_h_dR) */
static double J_h_Xr[9], J_h_dR[6];
/* tableaux auxilliaires*/
static double aux[9]; 
static double V_dep[3];

/*
 * Little bits of calife's libmat copied here to avoid linking with it
 *
 * XXXXX should be replaced some day...
 */
static int dms_plus(const double *m1, const double *m2, double *m_res, 
		int l, int c);
static int dmfsf_jmult2(const double *m1, const double *m2, double *m_res, 
		int l1, int c1, int l2, int c2);


#define MATELT(mat,i,j,c)   ( *((mat) + (c) * (i) + (j)) )
#define INDEX_SMATELT(i,j)   \
((i)<(j) ? ((j)*((j)+1))/2 + (i) : ((i)*((i)+1))/2 +(j)) 
#define SMATELT(mat,i,j,n) *((mat) + INDEX_SMATELT((i),(j)))
#define MATSIZE(l,c)        ((l) * (c))
#define SMATSIZE(n,c)       (((n) * ((n) + 1)) / 2)

static int 
dms_plus(const double *m1, const double *m2, double *m_res, int l, int c)
{
	int k;

	if (l != c)
	{
		fprintf (stderr, "dms : matrice symetrique non carree !!!\n");
		return -1;
	}

	for (k = 0; k < SMATSIZE(l, c); k++)
		*m_res++ = *m1++ + *m2++;
	return 0;
}

static int
dmfsf_jmult2 (const double *m1, const double *m2, double *m_res,
		int l1, int c1, int l2, int c2)
{

	double *temp, *res;
	double *aux = NULL;
	double *aux2 = NULL;
	int i,j,k,flag = 0;
	double val;

	if (l2 != c2 || c1 != c2)
	{
		fprintf(stderr, "%s : tailles incorrectes \n", __func__);
		return -1;
	}

	if (m_res == m1 || m_res == m2)
	{
		temp = malloc((l1*c2 + l1*(l1+1)/2) * sizeof(double));
		if (temp == NULL) {
			fprintf(stderr, "%s : Can't allocate memory for temp\n", __func__);
			return -1;
		}
		aux = temp;
		res = temp + l1*c2;
		aux2 = res;
		flag = 1;
	}
	else
	{
		temp = malloc((l1 * c2) * sizeof(double));
		if (temp == NULL) {
			fprintf(stderr, "%s : Can't allocate memory for temp\n", __func__);
			return -1;
		}
		aux = temp;
		res = m_res;
	}


	for (i = 0; i < l1; i++)
		for (j = 0; j < c2; j++)
		{
			val = 0.;
			for (k = 0; k < c1; k++)
				val += MATELT (m1, i, k, c1) * SMATELT (m2, k, j, c2);
			*temp++ = val;
		}
	temp = aux;

	for (j = 0; j < l1; j++)
		for (i = 0; i <= j; i++)
		{
			val = 0.;
			for (k = 0; k < c2; k++)
				val += MATELT (temp, i, k, c2) * MATELT (m1, j, k, c1);
			*res++ = val;
		}

	if (flag)
	{
		res = aux2;
		for (i = 0; i < SMATSIZE (l1, l1); i++)
			*m_res++ = *res++;
	}

	free(temp);

	return 0;
}

/*----------------------------------------------------------------------*/

/**
 ** fonction "h" de l'odometrie + jacobiennes
 **/

static void 
calc_nouv_pos(JLOCO_ODO *posId, double *jxr, double *jdep,  double LaX)
{
	double s, c;
	double dS, dTheta, thetaMoy;
	
	/*
	 * nouvelle position; equation :
	 * x = x + (dRl + dRr)/2 *cos(theta + dtheta/2);
	 * y = y + (dRl + dRr)/2 *sin(theta + dtheta/2);
	 * theta = theta + (dRr - dRl)/L ;
	 * 
	 * dtheta = (dRr - dRl)/L
	 * 
	 * L = longueur de l'entre-axe
	 * 
	 * Bon, ca c'est approximatif: a verifier. En tous cas, 
	 * une fois ecrite, il faut calculer les jacobiennes 
	 * suivant le modele qui suit
	 * 
	 */
	
	/* Calculer et stocker le deplacement curviligne du robot */
	posId->robotMisc.dS = dS = (posId->rightWheel.dS 
	    + posId->leftWheel.dS)/2.;
	posId->robotMisc.s += dS;
	
	/* Calculer et garder la variation d'orientation du robot */
	posId->robotMisc.dTheta = dTheta
	    = (posId->rightWheel.dS - posId->leftWheel.dS) / LaX;
	
	thetaMoy = posId->robot.theta + dTheta / 2;
	
	c = cos(thetaMoy);
	s = sin(thetaMoy);
	
	/* nouvelle position */
	posId->robot.xRob += dS * c;
	posId->robot.yRob += dS * s;
	posId->robot.theta = angleLimit(posId->robot.theta + dTheta);
	
	/* Calculer le sinus et le cosinus de cet angle et les stocker */
	posId->robotMisc.sinTheta = sin (posId->robot.theta);
	posId->robotMisc.cosTheta = cos (posId->robot.theta);
	
	/* Calculer et stocker les vitesses lineaire et angulaire du robot */
	posId->robot.v = (posId->rightWheel.v + posId->leftWheel.v) / 2.;
	posId->robot.w = (posId->rightWheel.v - posId->leftWheel.v) / LaX;
	
	/* Enfin, calculer le rayon de courbure */
	if (fabs (dTheta) < JLOCO_INFINY_CURV_DTHETA)
		posId->robotMisc.cR = JLOCO_INFINY_CURV_RADIUS;
	else posId->robotMisc.cR = dS / dTheta;
	
	/* jacobiennes */
	
	/* premiere ligne jxr */
	jxr[0] = 1.0;
	jxr[1] = 0;
	jxr[2] = -dS * s;
	/* deuxieme ligne jxr */
	jxr[3] = 0;
	jxr[4] = 1.0;
	jxr[5] = dS * c;
	/* troisieme ligne jxr */
	jxr[6] = 0;
	jxr[7] = 0;
	jxr[8] = 1.0;
	
	/* premiere ligne de jdep */
	jdep[0] = (c + dS*s/LaX)/2.0;
	jdep[1] = (c - dS*s/LaX)/2.0;
	/* deuxieme ligne de jdep */
	jdep[2] = (s - dS*c/LaX)/2.0;
	jdep[3] = (s + dS*c/LaX)/2.0;
	/* troisieme ligne de jdep */
	jdep[4] = -1/LaX;
	jdep[5] = 1/LaX;
	
	/* bon ca doit etre bon car c'est maple qui l'a calcul... */
	/* cf directory loco/maple and book : "Automous Mobile Robots",
	   Siegwart and Nourbakhsh, MIT press 2004, pp 187-189 */
}


/* variance sur le deplacement des roues; je suppose qu'il existe un
   facteur commun de proportion sur le deplacement; peut etre pas vrai;
   je pense qu'un veritable modele devrait contenir un terme constant
   plus un terme proportionel (quoi que si le robot ne bouge pas ... le
   terme constant devient plus que discutable !) 

   Il faudrait surtout un terme proportionnel a l'acceleration 
   (source de glissement)
*/

static void 
Calc_var_whee(JLOCO_ODO *posId, double *vdep, double coeff)
{
	/* this matrix is symmetric : 
	   dS_left    0     
	   0    dS_right
	*/
	vdep[0] = coeff * posId->leftWheel.dS * posId->leftWheel.dS;
	vdep[1] = 0;
	vdep[2] = coeff * posId->rightWheel.dS * posId->rightWheel.dS;
}


/*----------------------------------------------------------------------*/
/**
 ** fonction de calcul de l'odometrie
 **
 ** entree: posId - ancienne position + incertitude 
 **                 deplacements mesures
 **
 ** sortie: posId - nouvelle position + incertitude
 **/

void 
odoProba0(JLOCO_ODO *posId,  	/* ident poster Odometrie */
    double lax, 	/* entre-axe */
    double coeff)
{

	/*
	 * calcul de la nouvelle position et des jacobiennes en fonction des
	 * deplacements des roues.
	 */
	calc_nouv_pos(posId, J_h_Xr, J_h_dR, lax);
	
	/*
	 * variance sur le deplacement
	 */
	Calc_var_whee(posId, V_dep, coeff);
	
	/*
	 * calcul de la variance sur Xrob
	 */
	dmfsf_jmult2(J_h_dR, V_dep, aux, 3, 2, 2, 2);
	dmfsf_jmult2(J_h_Xr, posId->odoError.var, posId->odoError.var, 
	    3, 3, 3, 3);
	dms_plus(aux, posId->odoError.var, posId->odoError.var, 3, 3);
	
}


/*----------------------------------------------------------------------*/

/**
 ** fonction de calcul de l'odometrie + gyroscope
 **
 ** entree: odoId  - ancienne position + incertitude 
 **                  deplacements mesures: dS et v
 **         gyroId - ancienne position + incertitude 
 **                  deplacements mesures: theta, dTheta
 **
 ** sortie: gyroId - nouvelle position + incertitude
 **/

void odoGyroProba(JLOCO_ODO *odoId, 	/* id Odometrie */
		  JLOCO_GYRO *gyroId,  	/* id Gyroscope */
		  double lax)	        /* entre-axe */
{
  double thetaMoy;          /* Angle moyen entre 2 iterations */
  double dS;                /* Deplacement curviligne */
  double dT;                /* Variation angulaire */
  double theta;
  int i;

  /* Angle par gyroscope (limite entre -Pi et Pi) */
  dT = gyroId->robotMisc.dTheta;
  theta = angleLimit (gyroId->robot.theta + dT);
  gyroId->robot.theta = theta;
  gyroId->robotMisc.sinTheta = sin (theta);
  gyroId->robotMisc.cosTheta = cos (theta);
  
  /* Vitesse et variation curviligne par odometrie */
  dS = odoId->robotMisc.dS;
  gyroId->robotMisc.dS = dS;
  gyroId->robotMisc.s += dS;
  gyroId->robot.v = odoId->robot.v;

  /* Calculer la position du centre du robot */
  thetaMoy = theta - dT/2.;
  gyroId->robot.xRob += dS * cos(thetaMoy);
  gyroId->robot.yRob += dS * sin(thetaMoy);

  /* Enfin, calculer le rayon de courbure */
  if (fabs (dT) < JLOCO_INFINY_CURV_DTHETA)
    gyroId->robotMisc.cR = JLOCO_INFINY_CURV_RADIUS;
  else gyroId->robotMisc.cR = dS / dT;

  /* Proba: Pour le moment on reprend celle de l'odo ce qui est faux !! XXXX */
  for (i=0; i<6; i++)
    gyroId->odoError.var[i] = odoId->odoError.var[i];
  
}



